Gladius = LibStub("AceAddon-3.0"):NewAddon("Gladius", "AceEvent-3.0", "AceConsole-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Gladius", true)
local db

local arenaUnits, unitsToIdentify = {}, {}

function Gladius:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("GladiusDB", self:getDefaults())
	self:SetupOptions()
	db = self.db.profile
	self.buttons = {}
	self.currentBracket = nil
	
	--default arena units
	for i=1, 5 do
		arenaUnits[i] = "arena" .. i
		arenaUnits[arenaUnits[i]] = true
	end

end

function Gladius:OnEnable()
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "ZONE_CHANGED_NEW_AREA")
	if (not db.locked and db.x == 0 and db.y == 0) then
		self:ToggleFrame(5)
		self:Print(L["Welcome to Gladius!"])
		self:Print(L["First run has been detected, displaying test frame."])
		self:Print(L["Valid slash commands are:"])
		self:Print(L["/gladius ui"])
		self:Print(L["/gladius test1-5"])	
		self:Print(L["If this is not your first run please lock or move the frame to prevent this from happening."])
	end
end

function Gladius:ZONE_CHANGED_NEW_AREA()
	local type = select(2, IsInInstance())
		
	if (type == "arena") then
		self:JoinedArena()
	elseif (type ~= "arena" and instanceType == "arena") then
		self:LeftArena()
	end
	
	instanceType = type
end

function Gladius:ClearAllUnits()
	for k, v in pairs(self.buttons) do
		v:Hide()
	end
	
	for i=1, 5 do
		table.remove(self.buttons, i)
	end
end

--We joined an arena: clear units, register events, create buttons and show frames
function Gladius:JoinedArena()
	self:ClearAllUnits()
		
	self:RegisterEvent("UNIT_AURA")
	self:RegisterEvent("UNIT_HEALTH")
	self:RegisterEvent("UNIT_MANA", "UNIT_POWER")
	self:RegisterEvent("UNIT_RAGE", "UNIT_POWER")
	self:RegisterEvent("UNIT_ENERGY", "UNIT_POWER")
	self:RegisterEvent("UNIT_FOCUS", "UNIT_POWER")
	self:RegisterEvent("UNIT_RUNIC_POWER", "UNIT_POWER")
	self:RegisterEvent("UNIT_DISPLAYPOWER")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	
	self:RegisterEvent("UNIT_SPELLCAST_START")
	self:RegisterEvent("UNIT_SPELLCAST_STOP")
	self:RegisterEvent("UNIT_SPELLCAST_DELAYED")
	self:RegisterEvent("UNIT_SPELLCAST_FAILED", "UNIT_SPELLCAST_STOP")
	self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED", "UNIT_SPELLCAST_STOP")
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START")
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE", "UNIT_SPELLCAST_DELAYED")
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP", "UNIT_SPELLCAST_STOP")
	
	for i=1, MAX_BATTLEFIELD_QUEUES do
		local status, _, _, _, _, teamSize = GetBattlefieldStatus(i)
		if (status == "active" and teamSize > 0) then
			Gladius.currentBracket = teamSize
			break
		end
	end
	
	for i=1, Gladius.currentBracket do
		local unit = arenaUnits[i]
		local button = self.buttons[unit]
		if (not button) then
			button = self:CreateButton(i)
			self.buttons[unit] = button
		end
		
		button.isDead = nil
		button.powerType = 0
		table.insert(unitsToIdentify, unit)	
	end
	
	self.frame.testing = false
	self:UpdateFrame()
	self.UpdateUnits()
	self:UpdateBindings()
	self.frame:Show()
	self.scanFrame:Show()
end

--We left an arena, unregister events and clear/hide frames
function Gladius:LeftArena()
	self:UnregisterAllEvents()
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "ZONE_CHANGED_NEW_AREA")
	self:ClearAllUnits()	
	self.frame:Hide()
	self.scanFrame:Hide()
end

--Update units health
function Gladius:UNIT_HEALTH(event, unit)
	if (arenaUnits[unit] and not self.buttons[unit].isDead) then
		local health = math.floor((UnitHealth(unit)/UnitHealthMax(unit)) * 100)
		local button = self.buttons[unit]
		button.health:SetValue(health)
		button.healthText:SetText(health.."%")
	end
end

--Update units mana/rage/energy
function Gladius:UNIT_POWER(event, unit)
	if (arenaUnits[unit] and not self.buttons[unit].isDead) then	
		local mana = UnitMana(unit)
		local manaMax = UnitManaMax(unit)
		local manaPercentage = math.floor((UnitMana(unit)/UnitManaMax(unit)) * 100)
		local button = self.buttons[unit]
		
		button.mana:SetValue(manaPercentage)
		button.manaText:SetText("")
		
		if db.manaActual then
			button.manaText:SetText(string.format("%d", mana))
		end
		
		if db.manaMax then
			if button.manaText:GetText() then
				button.manaText:SetText(string.format("%s%s%d", button.manaText:GetText(), "/", manaMax))
			else
				button.manaText:SetText(string.format("%d", manaMax))
			end
		end
		
		if db.manaPercentage then
			if button.manaText:GetText() then
				button.manaText:SetText(string.format("%s %s%d%s", button.manaText:GetText(), "(", manaPercentage, "%)"))
			else
				button.manaText:SetText(string.format("%d%s", manaPercentage, "%"))
			end		
		end
		
	end
end

--Update units powertypes
function Gladius:UNIT_DISPLAYPOWER(event, unit)
	if (arenaUnits[unit] and not self.buttons[unit].isDead) then
		local button = self.buttons[unit]
		button.powerType = UnitPowerType(unit)
		if (button.powerType == 0 and not db.manaDefault) then
			button.mana:SetStatusBarColor(db.manaColor.r, db.manaColor.g, db.manaColor.b, db.manaColor.a)
		elseif (button.powerType == 1 and not db.rageDefault) then
			button.mana:SetStatusBarColor(db.rageColor.r, db.rageColor.g, db.rageColor.b, db.rageColor.a)
		elseif (button.powerType == 3 and not db.energyDefault) then
			button.mana:SetStatusBarColor(db.energyColor.r, db.energyColor.g, db.energyColor.b, db.energyColor.a)
		else
			button.mana:SetStatusBarColor(PowerBarColor[button.powerType].r, PowerBarColor[button.powerType].g, PowerBarColor[button.powerType].b)
		end
		self:UNIT_POWER(nil, unit)
	end
end

--Auras
function Gladius:UNIT_AURA(event, unit)
	if (arenaUnits[unit] and not self.buttons[unit].isDead) then
		local aura = self.buttons[unit].auraFrame
		local index = 1
		while (true) do
			local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitAura(unit, index, "HELPFUL")
			if (not name) then break end
			index = index+1
			local priority
			if (GladiusAuraList[name]) then
				priority = GladiusAuraList[name]
				if (not aura.auraActive)  then
					Gladius:AuraGain(unit, name, icon, expirationTime, priority)
				else
					if (priority >= aura.priority) then
						Gladius:AuraGain(unit, name, icon, expirationTime, priority)
					end
				end
			end	
		end
		
		index = 1
		while (true) do
			local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitAura(unit, index, "HARMFUL")
			if (not name) then break end
			index = index+1
			local priority
			if (GladiusAuraList[name]) then
				priority = GladiusAuraList[name]
				if (not aura.auraActive)  then
					Gladius:AuraGain(unit, name, icon, expirationTime, priority)
				else
					if (priority >= aura.priority) then
						Gladius:AuraGain(unit, name, icon, expirationTime, priority)
					end
				end
			end	
		end
	end
end

function Gladius:AuraGain(unit, name, icon, expirationTime, priority)
	if (not db.auras) then return end
	local aura = self.buttons[unit].auraFrame
	aura.name = name
	aura.priority = priority
	aura.timeLeft = (expirationTime - GetTime())
	aura.icon:SetTexture(icon)
	aura.auraActive = true
end

function Gladius:AuraFades(bar)
	bar.name = nil
	bar.priority = nil
	bar.text:SetText("")
	bar.icon:SetTexture("")
	bar.auraActive = nil
end

--Spell casts
function Gladius:UNIT_SPELLCAST_START(event, unit)
	if (arenaUnits[unit] and not self.buttons[unit].isDead) then
		local spell, rank, displayName, icon, startTime, endTime, isTradeSkill = UnitCastingInfo(unit)
		local castBar = self.buttons[unit].castBar
		castBar.isCasting = true
		castBar.value = (GetTime() - (startTime / 1000))
		castBar.maxValue = (endTime - startTime) / 1000
		castBar:SetMinMaxValues(0, castBar.maxValue)
		castBar:SetValue(castBar.value)
		castBar.timeText:SetText(maxValue)
		castBar.spellText:SetText(spell .. " (" .. rank .. ")")
	end
end

function Gladius:UNIT_SPELLCAST_CHANNEL_START(event, unit)
	if (arenaUnits[unit] and not self.buttons[unit].isDead) then
		local spell, rank, displayName, icon, startTime, endTime, isTradeSkill = UnitChannelInfo(unit)
		local castBar = self.buttons[unit].castBar
		castBar.isChanneling = true
		castBar.value = ((endTime / 1000) - GetTime())
		castBar.maxValue = (endTime - startTime) / 1000
		castBar:SetMinMaxValues(0, castBar.maxValue)
		castBar:SetValue(castBar.value)
		castBar.timeText:SetText(maxValue)
		castBar.spellText:SetText(spell .. " (" .. rank .. ")")
	end	
end

function Gladius:UNIT_SPELLCAST_STOP(event, unit)
	if (arenaUnits[unit] and not self.buttons[unit].isDead) then
		self:CastEnd(self.buttons[unit].castBar)
	end
end

function Gladius:UNIT_SPELLCAST_DELAYED(event, unit)
	if (arenaUnits[unit] and not self.buttons[unit].isDead) then
		local spell, rank, displayName, icon, startTime, endTime, isTradeSkill
		if (event == "UNIT_SPELLCAST_DELAYED") then
			spell, rank, displayName, icon, startTime, endTime, isTradeSkill = UnitCastingInfo(unit)
		else
			spell, rank, displayName, icon, startTime, endTime, isTradeSkill = UnitChannelInfo(unit)
		end
		local bar = self.buttons[unit].castBar
		bar.value = (GetTime() - (startTime / 1000))
		bar.maxValue = (endTime - startTime) / 1000
		bar:SetMinMaxValues(0, bar.maxValue)
	end
end

function Gladius:CastEnd(bar)
	bar.isCasting = nil
	bar.isChanneling = nil
	bar.timeText:SetText("")
	bar.spellText:SetText("")
	bar:SetValue(0)
end

--Update the highlight and frame border for current target
function Gladius:PLAYER_TARGET_CHANGED(event)
	local target = UnitGUID("target")
	for k, v in pairs(self.buttons) do
		if (v.GUID == target) then
			if (db.highlight) then
				v.highlight:Show()
			end
			if (db.selectedBorder) then
				v.selected:SetBackdropBorderColor(db.selectedFrameColor.r,db.selectedFrameColor.g,db.selectedFrameColor.b,db.selectedFrameColor.a)
			end
		else
			v.highlight:Hide()
			v.selected:SetBackdropBorderColor(db.selectedFrameColor.r,db.selectedFrameColor.g,db.selectedFrameColor.b,0)
		end
	end
end

--Scan combatlog for enemy deaths and aura fading
function Gladius:COMBAT_LOG_EVENT_UNFILTERED(event, timestamp, eventType, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	if (eventType == "PARTY_KILL") then
		self:EnemyDeath(destGUID)
	elseif (eventType == "SPELL_AURA_REMOVED" or eventType == "SPELL_PERIODIC_AURA_REMOVED" or eventType == "SPELL_AURA_REMOVED_DOSE" or eventType == "SPELL_PERIODIC_AURA_REMOVED_DOSE") then
		local spellId, spellName = ...
		for i=1, 5 do
			local unit = arenaUnits[i]
			if (UnitGUID(unit) == destGUID and spellName == self.buttons[unit].auraFrame.name) then
				self:AuraFades(self.buttons[unit].auraFrame)
			end
		end
	end
end

--Enemy has died
function Gladius:EnemyDeath(GUID)
	for i=1, 5 do
		local unit = arenaUnits[i]
		if (UnitGUID(unit) == GUID) then
			local button = self.buttons[unit]
			button.health:SetValue(0)
			button.healthText:SetText("0%")
			button.mana:SetValue(0)
			button.manaText:SetText("0%")
			button.classIcon:SetAlpha(0.33)
			button.isDead = true
		end
	end
end


--Update single unit button
function Gladius:UpdateUnit(unit)	
		local button = self.buttons[unit]
		--check if unit exists
		if (UnitExists(unit)) then
			local name, server = UnitName(unit)
			local classLoc, class = UnitClass(unit)
			local raceLoc, race = UnitRace(unit)
			local sex = UnitSex(unit)
			
			button.name = name
			button.GUID = UnitGUID(unit)
			button.text:SetText(name)
			button.classText:SetText(raceLoc)
					
			if (not db.classText) then
				button.classText:Hide()
			end
						
			--health bar class color
			button.health:SetStatusBarColor(RAID_CLASS_COLORS[class].r, RAID_CLASS_COLORS[class].g, RAID_CLASS_COLORS[class].b, 1)
			
			--class icon
			local coords = CLASS_BUTTONS[class]
			button.classIcon:SetTexture("Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes")
			button.classIcon:SetTexCoord(coords[1], coords[2], coords[3], coords[4])
			button.classIcon:SetAlpha(1)
			
			--Update selected frame
			if (button.GUID == UnitGUID("target") and db.selectedBorder) then
				button.selected:SetBackdropBorderColor(db.selectedFrameColor.r,db.selectedFrameColor.g,db.selectedFrameColor.b,db.selectedFrameColor.a)
			elseif (button.GUID ~= UnitGUID("target") or (button.GUID == UnitGUID("target") and not db.selectedBorder)) then
				button.selected:SetBackdropBorderColor(db.selectedFrameColor.r,db.selectedFrameColor.g,db.selectedFrameColor.b,0)
			end
			
			-- Update health/mana/power
			self:UNIT_HEALTH(nil, unit)
			self:UNIT_POWER(nil, unit)
			self:UNIT_DISPLAYPOWER(nil, unit)
			self:UNIT_AURA(nil, unit)
			self:CastEnd(button.castBar)
			
			button:SetAlpha(1)		
		else
			button:SetAlpha(0)
		end
		
		--Attributes for targeting and setting focus
		self:UpdateAttributes(unit)
							
		--show button
		button:Show()
		
end

--Update all buttons that are used in current arena
function Gladius:UpdateUnits()
	local currentBracket = Gladius.currentBracket
	for i=1, currentBracket do
		Gladius:UpdateUnit("arena" .. i)
	end
end

--Update the button bindings
function Gladius:UpdateBindings()
	for k, v in pairs(self.buttons) do
		local keyLeft = GetBindingKey("GLADIUSTARGET"..v.id.."_LEFT")
		local keyRight = GetBindingKey("GLADIUSTARGET"..v.id.."_RIGHT")
		
		if (keyLeft) then
			SetOverrideBindingClick(v.secure, false, keyLeft, v.secure:GetName(), "LeftButton")
		elseif (keyRight) then
			SetOverrideBindingClick(v.secure, false, keyRight, v.secure:GetName(), "RightButton")		
		elseif (not keyLeft and not keyRight) then
			ClearOverrideBindings(v.secure)
		end		
		
	end
end

--Update attribute
function Gladius:UpdateAttribute(unit, key, action, spellName)
	local button = self.buttons[unit].secure
	local test = self.frame.testing
	
	if (test and unit == "arena1") then
		button:SetAttribute("unit", "player")
	else
		button:SetAttribute("unit", unit)
	end

	if (action ~= "spell" and action ~= "disabled") then
		button:SetAttribute("type" .. key, action)
	elseif (action == "spell" and spellName ~= "") then
		button:SetAttribute("type" .. key, "spell")
		button:SetAttribute("spell" .. key, spellName)
	end
end

--Update all attributes for a specific unit
function Gladius:UpdateAttributes(unit)
	self:UpdateAttribute(unit, 1, db.leftClick, db.leftSpell)
	self:UpdateAttribute(unit, 2, db.rightClick, db.rightSpell)
	self:UpdateAttribute(unit, 3, db.middleClick, db.middleSpell)
	self:UpdateAttribute(unit, 4, db.mouse4, db.mouse4spell)
	self:UpdateAttribute(unit, 5, db.mouse5, db.mouse5spell)
end

--Create an empty frame that scans for arena units and updates the corresponding button if unit exists
local scanFrame = CreateFrame("Frame")
scanFrame:Hide()

scanFrame:SetScript("OnUpdate", function()
		for i=#(unitsToIdentify), 1, -1 do
			local unit = unitsToIdentify[i]
			if (UnitExists(unit) and UnitName(unit) ~= UNKNOWNOBJECT) then
				Gladius:UpdateUnit(unit)
				table.remove(unitsToIdentify, i)
			end
		end
end)

Gladius.scanFrame = scanFrame


--Create five test buttons
function Gladius:Test()
	local currentBracket = Gladius.currentBracket
	for i=1, currentBracket do
		local unit = arenaUnits[i]
		local button = self.buttons[unit]
		if (not button) then
			button = self:CreateButton(i)
			self.buttons[unit] = button
		end
			
		button.isDead = nil
	
		local class, race, sex, classLoc, raceLoc, health, mana, manaMax, manaPercentage
		
		if (i == 1) then
			classLoc, class = UnitClass("player")
			raceLoc, race = UnitRace("player")
			sex = UnitSex("player")
			button.GUID = UnitGUID("player")
			button.name = UnitName("player")
			button.powerType = UnitPowerType("player")
			self:UpdateAttributes("arena1")
			health = math.floor((UnitHealth("player")/UnitHealthMax("player")) * 100)
			manaMax = UnitManaMax("player")
			mana = UnitMana("player")
			manaPercentage = math.floor((UnitMana("player")/UnitManaMax("player")) * 100)
		else
			class, race, sex = "HUNTER", "ORC", 2
			classLoc, raceLoc = "Hunter", "Orc"	
			button.name = "Arena " .. i
			button.GUID = "testframe"
			button.powerType = i-2
			manaMax = 6000
			health, mana = 100-(i^2), manaMax-(i^5)
			manaPercentage = math.floor((mana/manaMax) * 100)
		end
		
		button.manaMax = manaMax
		button.manaActual = mana
		button.manaPercentage = manaPercentage
		
		button.text:SetText(button.name)		
			
		button.castBar:SetMinMaxValues(0,i)
		button.castBar:SetValue(i-0.5)
		button.castBar.spellText:SetText("Example Spell (Rank 1)")
		button.castBar.timeText:SetText(i-0.5)
		
		Gladius:AuraGain("arena1", "Ice Block", select(3, GetSpellInfo(45438)), GetTime()+10, 3)
		
		button.health:SetValue(health)
		button.healthText:SetText(health.."%")
		button.mana:SetValue(manaPercentage)
		button.classText:SetText(raceLoc)
		
		if (not db.classText) then
			button.classText:Hide()
		end
		
		if (not db.manaText) then
			button.manaText:Hide()
		end
		
		if (button.powerType == 0 and not db.manaDefault) then
			button.mana:SetStatusBarColor(db.manaColor.r, db.manaColor.g, db.manaColor.b, db.manaColor.a)
		elseif (button.powerType == 1 and not db.rageDefault) then
			button.mana:SetStatusBarColor(db.rageColor.r, db.rageColor.g, db.rageColor.b, db.rageColor.a)
		elseif (button.powerType == 3 and not db.energyDefault) then
			button.mana:SetStatusBarColor(db.energyColor.r, db.energyColor.g, db.energyColor.b, db.energyColor.a)
		else
			button.mana:SetStatusBarColor(PowerBarColor[button.powerType].r, PowerBarColor[button.powerType].g, PowerBarColor[button.powerType].b)
		end
		
		button.health:SetStatusBarColor(RAID_CLASS_COLORS[class].r, RAID_CLASS_COLORS[class].g, RAID_CLASS_COLORS[class].b, 1.0)
			
		local coords = CLASS_BUTTONS[class]
		button.classIcon:SetTexture("Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes")
		button.classIcon:SetTexCoord(coords[1], coords[2], coords[3], coords[4])
		button.classIcon:SetAlpha(1)
			
		button:Show()
	end
end

